/*
	This simple demo program initializes an SSD1306 OLED display, connected
	via the i2c bus.
*/

#define _GNU_SOURCE
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "i2c-dev.h"
#include "ssd1306.h"

#define DISPLAY_ADDRESS 0x3c

extern const unsigned char char_set[];
static unsigned char curr_x, curr_y;

static unsigned char buffer[1024];
static void displayString(const char *string);
static void displayUpdate(void);

static int fd;
static int pokestream8(int twifd, const unsigned char *data, unsigned char reg, int size);
static void poke8(int twifd, unsigned char reg, unsigned char data);

const unsigned char initdata[]  = {
	SSD1306_SETCONTRAST, 0xCF,
	SSD1306_SEGREMAP | 1,
	SSD1306_COMSCANDEC,
	SSD1306_SETMULTIPLEX, SSD1306_LCDHEIGHT - 1,
	SSD1306_SETDISPLAYOFFSET, 0,
	SSD1306_SETSTARTLINE,
	SSD1306_SETDISPLAYCLOCKDIV, 0x80,
	SSD1306_SETPRECHARGE, 0xF1,
	SSD1306_SETCOMPINS, 0x12,
	SSD1306_SETVCOMDETECT, 0x40,
	SSD1306_CHARGEPUMP, 0x14,
	SSD1306_MEMORYMODE, 0,
	SSD1306_COLUMNADDR, 0, SSD1306_LCDWIDTH - 1,
	SSD1306_PAGEADDR, 0, 7,
	SSD1306_DISPLAYALLON_RESUME,
	SSD1306_NORMALDISPLAY,
	SSD1306_DISPLAYON,
};

void ssd1306_Init(void)
{
	pokestream8(fd, initdata, 0, sizeof(initdata));
	poke8(fd, 0, SSD1306_COLUMNADDR);
	poke8(fd, 0, 0);   // Column start address (0 = reset)
	poke8(fd, 0, SSD1306_LCDWIDTH - 1); // Column end address (127 = reset)
	poke8(fd, 0, SSD1306_PAGEADDR);
	poke8(fd, 0, 0); // Page start address (0 = reset)
	poke8(fd, 0, 7); // Page end address

	memset(buffer, 0, sizeof(buffer));
	displayUpdate();
}


int main(int argc, char *argv[])
{
	if (argc < 2) {
		fprintf(stderr, "usage: %s i2c-device\n", 	program_invocation_short_name);
		return 1;
	}

	if ((fd = open(argv[1], O_RDWR)) < 0) {
		fprintf(stderr, "Can't open %s\n", argv[1]);
		return 1;
	}

	if (ioctl(fd, I2C_SLAVE_FORCE, DISPLAY_ADDRESS) < 0) {
		close(fd);
		fprintf(stderr, "No ACK from 0x%02X\n", DISPLAY_ADDRESS);
		return 1;
	}

	ssd1306_Init();
	displayString(" Hello, World!\n SSD1306 Demo\n   Over i2c");
	displayUpdate();

	close(fd);

	return 0;
}

static int pokestream8(int twifd, const unsigned char *data, unsigned char reg, int size)
{
	struct i2c_rdwr_ioctl_data packets;
	struct i2c_msg msg;
	unsigned char outdata[4096];

	/* Linux only supports 4k transactions at a time, and we need
	 * one bytes for the address, one for the reg */
	if (size > 4095)
		return 1;

	outdata[0] = reg;
	memcpy(&outdata[1], data, size);

	msg.addr = DISPLAY_ADDRESS;
	msg.flags = 0;
	msg.len	= 1 + size;
	msg.buf	= (char *)outdata;

	packets.msgs  = &msg;
	packets.nmsgs = 1;

	if(ioctl(twifd, I2C_RDWR, &packets) < 0) {
		return 1;
	}
	return 0;
}

static void poke8(int twifd, unsigned char reg, unsigned char data)
{
	int ret;
	ret = pokestream8(twifd, &data, reg, 1);
	if(ret) {
		perror("Failed to write to device");
	}
}

static void setPixel(unsigned char x, unsigned char y)
{
	if (x >= SSD1306_LCDWIDTH || y >= SSD1306_LCDHEIGHT)
		return;

	x = (SSD1306_LCDWIDTH-1) - x;
	y = (SSD1306_LCDHEIGHT-1) - y;

	buffer[x+ (y/8)*SSD1306_LCDWIDTH] |=  (1 << (y&7));

}


static void clrPixel(unsigned char x, unsigned char y)
{
	if (x >= SSD1306_LCDWIDTH || y >= SSD1306_LCDHEIGHT)
		return;

	x = (SSD1306_LCDWIDTH-1) - x;
	y = (SSD1306_LCDHEIGHT-1) - y;

	buffer[x+ (y/8)*SSD1306_LCDWIDTH] &= ~(1 << (y&7));

}


static void inline genchr(register int x, int y, int c)
{
	unsigned int line, max_line, start_x, max_x;
	register	const unsigned char *ptr;
	register	unsigned char chr;

	start_x = x;
	line = y;

	max_line = line + 19;
	max_x = x + 8;
	ptr = char_set;
	ptr += (c*19);

	for (line = y; line < max_line; line++) {
		chr = *ptr++;
		for (x = start_x; x < max_x; x++) {

			if (chr & 0x80) {
				setPixel(x,line);
			} else {
				clrPixel(x,line);
			}
			chr <<= 1;
		}
	}
}

static void displayPutc(int c)
{
	if (c == '\n') {
		curr_y += 19;
		curr_x = 0;
	} else if (c == '\r') {
		curr_x = 0;
	} else {
		genchr(curr_x, curr_y, c & 0xff);
		curr_x += 8;
	}
}

static void displayString(const char *string)
{
	while(*string)
		displayPutc(*string++);
}

static void displayUpdate(void)
{
	pokestream8(fd, buffer, 0x40, sizeof(buffer));
}
